import time
from functools import wraps
from typing import Optional

from loguru import logger


def format_time(seconds: float) -> str:
    hours = int(seconds // 3600)
    seconds %= 3600
    minutes = int(seconds // 60)
    seconds = int(seconds % 60)
    return f"{hours:02}:{minutes:02}:{seconds:02}"


def timing(custom_message: Optional[str] = None):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = func(*args, **kwargs)  # Call the original function
            end_time = time.perf_counter()
            execution_time = end_time - start_time
            message = f"{custom_message if custom_message is not None else ''} {func.__name__} took {format_time(execution_time)}"
            logger.debug(message)
            return result
        return wrapper
    return decorator


import time

from torch import Tensor
from transformers import AutoTokenizer, AutoModel


def average_pool(last_hidden_states: Tensor,
                 attention_mask: Tensor) -> Tensor:
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
    return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]


#def average_pool(hidden_state, attention_mask):
#    # Assuming you want to perform mean pooling over the non-padding part of the sequence
#    sum_hidden_state = torch.sum(hidden_state * attention_mask.unsqueeze(-1), dim=1)
#    count_hidden_state = torch.sum(attention_mask, dim=1).unsqueeze(-1)
#    return sum_hidden_state / count_hidden_state


def compute_embeddings_time(model_name: str, input_texts: list[str], cuda: bool = True):
    import torch

    if cuda and torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        print(f"trying to use cuda but it is not available!")
        device = torch.device('cpu')

    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModel.from_pretrained(model_name)
    model.to(device)

    total_execution_time = 0
    embeddings_list = []

    for text in input_texts:
        start_time = time.perf_counter()

        # Tokenize the input text
        batch_dict = tokenizer([text], max_length=512, padding=True, truncation=True, return_tensors='pt')
        batch_dict = {key: val.to(device) for key, val in batch_dict.items()}

        outputs = model(**batch_dict)
        embeddings = average_pool(outputs.last_hidden_state, batch_dict['attention_mask'])

        end_time = time.perf_counter()
        execution_time = end_time - start_time

        print(f"Execution time for text: {execution_time:.4f} seconds")
        total_execution_time += execution_time
        embeddings_list.append(embeddings)

    print(f"Total execution time: {total_execution_time:.4f} seconds")
    return torch.cat(embeddings_list, dim=0), total_execution_time




#!/usr/bin/env python3
from typing import Dict

import click
from click import Context
from hybrid_search.opensearch_hybrid_search import OpenSearchHybridSearch
from pycomfort.config import load_environment_keys, LOG_LEVELS, LogLevel, configure_logger
from qdrant_client import QdrantClient
from qdrant_client.http.models import PayloadSchemaType

from indexpaper.indexing import index_selected_papers, index_selected_documents, init_qdrant, fast_index_papers
from indexpaper.paperset import Paperset
from indexpaper.resolvers import *
from indexpaper.utils import timing
from opensearchpy import OpenSearch, exceptions

@click.group(invoke_without_command=False)
@click.pass_context
def app(ctx: Context):
    # if ctx.invoked_subcommand is None:
    #    click.echo('Running the default command...')
    #    test_index()
    pass


@app.command("index_papers")
@click.option('--papers', type=click.Path(exists=True), help="papers folder to index")
@click.option('--collection', default='papers', help='papers collection name')
@click.option('--folder', type=click.Path(), default=None, help="folder to put chroma indexes to (if we use chroma)")
@click.option('--url', type=click.STRING, default=None, help="alternatively you can provide url, for example http://localhost:6333 for qdrant")
@click.option('--key', type=click.STRING, default=None, help="your api key if you are using cloud vector store")
@click.option('--embeddings', type=click.Choice(EMBEDDINGS), default=EmbeddingType.OpenAI.value, help='size of the chunk for splitting')
@click.option('--chunk_size', type=click.INT, default=3000, help='size of the chunk in tokens for splitting (characters for recursive spliiter and tokens for openai one)')
@click.option("--model", type=click.Path(), default=None, help="path to the model (required for embeddings)")
@click.option('--include_meta', type=click.BOOL, default=True, help="if metadata is included")
@click.option('--database', type=click.Choice(VECTOR_DATABASES, case_sensitive=False), default=VectorDatabase.Chroma.value, help = "which store to take")
@click.option("--device", type=click.Choice(DEVICES), default=Device.cpu.value, help="which device to use, cpu by default, so do not forget to put cuda if you are using NVIDIA")
@click.option('--prefer_grpc', type=click.BOOL, default = None)
@click.option('--rewrite', type=click.BOOL, default=False, help = "Rewrite collection if it is already present")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def index_papers_command(papers: str, collection: str, folder: str, url: str, key: str, embeddings: str, chunk_size: int,  model: Optional[str], include_meta: bool, database: str, device: Optional[str], prefer_grpc: Optional[bool], rewrite: bool, log_level: str) -> Path:
    configure_logger(log_level)
    load_environment_keys(usecwd=True)
    papers_folder = Path(papers)
    assert not (folder is None and url is None and key is None), "either database folder or database url or api_key should be provided!"
    embedding_type = EmbeddingType(embeddings)
    splitter = resolve_splitter(embedding_type, model, chunk_size)
    return index_selected_papers(papers_folder, collection, splitter,  embedding_type, include_meta, folder, url,
                                 database=VectorDatabase[database],
                                 key=key,
                                 model=model,
                                 prefer_grpc=prefer_grpc,
                                 device=Device(device),
                                 always_recreate=rewrite
                                 )


@timing
@app.command("fast_index_papers")
@click.option('--papers', type=click.Path(exists=True), help="papers folder to index")
@click.option('--collection', default='dataset', help='dataset collection name')
@click.option('--url', type=click.STRING, required=True, help="URL or API key for example http://localhost:6333 for qdrant")
@click.option('--key', type=click.STRING, default=None, help="your api key if you are using cloud vector store")
@click.option("--model", type=click.Path(), default=EmbeddingModels.default.value, help="fast embedding model, BAAI/bge-base-en-v1.5 by default")
@click.option('--prefer_grpc', type=click.BOOL, default=True, help="only needed for qdrant database")
@click.option('--rewrite', type=click.BOOL, default=False, help = "Rewrite collection if it is already present")
@click.option('--paginated', type=click.BOOL, default=True, help = "If it is paginated paper")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def fast_index_papers_command(papers: str, collection: str, url: Optional[str], key: Optional[str], model: str, prefer_grpc: bool, rewrite: bool, paginated: bool, log_level: str) -> Path:
    configure_logger(log_level)
    return fast_index_papers(Path(papers), collection, url, key, model, prefer_grpc, rewrite, paginated)


#@timing
@app.command("hybrid_index")
@click.option('--dataset', type=click.STRING, help="Dataset to index, can be either Path or hugging face dataset")
@click.option('--collection', required=True, help='dataset collection name')
@click.option('--url', type=click.STRING, required=False, help="URL for opensearch, http://localhost:9200 by default for OpenSearch")
@click.option("--model", type=click.Path(), default=EmbeddingModels.default.value, help="fast embedding model, BAAI/bge-base-en-v1.5 by default")
@click.option("--device", type=click.Choice(DEVICES), default=Device.cpu.value, help="which device to use, cpu by default, so do not forget to put cuda if you are using NVIDIA")
@click.option('--start', type=click.INT, default=0, help='When to start slicing the dataset')
@click.option('--content_field', type=click.STRING, default="annotations_paragraph", help = "default dataset content field")
@click.option('--paragraphs', type=click.INT, default=5, help='number of paragraphs to connect together when preprocessing')
@click.option('--slice', type=click.INT, default=100, help='What is the size of the slice')
@click.option('--chunk_size', type=click.INT, default=512, help='What is the size of the chunk')
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def hybrid_index_command(dataset: str, collection: str, url: Optional[str], model: str, device: str, start: int, content_field: str, paragraphs: int, slice: int, chunk_size: int, log_level: str) -> Path:
    logger = configure_logger(log_level)
    load_environment_keys(usecwd=True)
    logger.add("./logs/hybrid_index_{time}.log")
    extras = {
        "dataset": str(dataset),
        "paragraphs_number": str(paragraphs),
        "model": str(model)
    }
    logger.info(f"computing embeddings into collection {collection} for {dataset} with model {model} using slices of {slice} starting from {start} with chunks of {chunk_size} tokens when splitting")
    splitter = HuggingFaceSplitter(model, tokens=chunk_size)
    paper_set = Paperset(dataset, splitter=splitter, content_field=content_field, paragraphs_together=paragraphs)
    embeddings = resolve_embeddings(EmbeddingType.HuggingFaceBGE if "bge" in model else EmbeddingType.HuggingFace, model = model, device = Device(device))
    logger.info(f"computing embeddings into collection {collection} for {dataset} with model {model} using slices of {slice} starting from {start} with chunks of {chunk_size} tokens when splitting")
    url = os.environ.get("OPENSEARCH_URL") if url is None else url
    login = os.getenv("OPENSEARCH_USER", "admin")
    password = os.getenv("OPENSEARCH_PASSWORD", "admin")
    logger.info(f"initializing opensearch at {url}")
    hybrid: OpenSearchHybridSearch = OpenSearchHybridSearch.create(url, collection, embeddings, login=login, password=password)
    #if not hybrid.check_pipeline_exists():
    #    logger.info(f"hybrid search pipeline does not exist, creating it for {url}")
    #    hybrid.create_pipeline(url)
    result = paper_set.index_hybrid_by_slices(slice, hybrid, start, logger=logger, extras = extras)
    return result

import click
from typing import Optional, List

# Assume all the necessary imports and existing command definitions are already in place

@app.command("hybrid_index_multiple")
@click.option('--datasets', multiple=True, type=click.Path(exists=True), help="Datasets to index, can be either Paths or hugging face datasets")
@click.option('--collection', required=True, help='dataset collection name')
@click.option('--url', type=click.STRING, required=False, help="URL for opensearch, http://localhost:9200 by default for OpenSearch")
@click.option("--model", type=click.Path(), default=EmbeddingModels.default.value, help="fast embedding model, BAAI/bge-base-en-v1.5 by default")
@click.option("--device", type=click.Choice(DEVICES), default=Device.cpu.value, help="which device to use, cpu by default, so do not forget to put cuda if you are using NVIDIA")
@click.option('--content_field', type=click.STRING, default="annotations_paragraph", help = "default dataset content field")
@click.option('--paragraphs', type=click.INT, default=5, help='number of paragraphs to connect together when preprocessing')
@click.option('--slice', type=click.INT, default=100, help='What is the size of the slice')
@click.option('--chunk_size', type=click.INT, default=512, help='What is the size of the chunk')
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
@click.pass_context
def hybrid_index_multiple_command(ctx, datasets: List[str], collection: str, url: Optional[str], model: str, device: str, content_field: str, paragraphs: int, slice: int, chunk_size: int, log_level: str):
    logger = configure_logger(log_level)
    load_environment_keys(usecwd=True)
    logger.add("./logs/hybrid_index_multiple_{time}.log")
    for dataset in datasets:
        try:
            ctx.invoke(hybrid_index_command, dataset=dataset, collection=collection, url=url, model=model, device=device, start=0, content_field=content_field, paragraphs=paragraphs, slice=slice, chunk_size=chunk_size, log_level=log_level)
            logger.info(f"Successfully processed dataset: {dataset}")
        except Exception as e:
            logger.error(f"Error processing dataset {dataset}: {e}")
            # Continue to the next dataset without stopping the command execution


@timing
@app.command("fast_index")
@click.option('--dataset', type=click.STRING, help="Dataset to index, can be either Path or hugging face dataset")
@click.option('--collection', default='dataset', help='dataset collection name')
@click.option('--url', type=click.STRING, required=True, help="URL or API key for example http://localhost:6333 for qdrant")
@click.option('--key', type=click.STRING, default=None, help="your api key if you are using cloud vector store")
@click.option("--model", type=click.Path(), default=EmbeddingModels.default.value, help="fast embedding model, BAAI/bge-base-en-v1.5 by default")
@click.option('--prefer_grpc', type=click.BOOL, default=False, help = "only needed for qdrant database")
@click.option('--start', type=click.INT, default=0, help='When to start slicing the dataset')
@click.option('--content_field', type=click.STRING, default="annotations_paragraph", help = "default dataset content field")
@click.option('--paragraphs', type=click.INT, default=5, help='number of paragraphs to connect together when preprocessing')
@click.option('--slice', type=click.INT, default=100, help='What is the size of the slice')
@click.option('--batch_size', type=click.INT, default=50, help="Batch size, 50 by default")
@click.option('--parallel', type=click.INT, default=10, help="How many workers to use")
@click.option('--rewrite', type=click.BOOL, default=False, help = "Rewrite collection if it is already present")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def fast_index_command(dataset: str, collection: str, url: Optional[str], key: Optional[str], model: str, prefer_grpc: bool, start: int, content_field: str, paragraphs: int, slice: int, batch_size: int, parallel: Optional[int], rewrite: bool, log_level: str) -> Path:
    configure_logger(log_level)
    load_environment_keys(usecwd=True)
    assert not (url is None and key is None), "either database url or api_key should be provided!"
    chunk_size: int = 512
    logger.info(f"computing embeddings into collection {collection} for {dataset} with model {model} using slices of {slice} starting from {start} with chunks of {chunk_size} tokens when splitting")
    splitter = HuggingFaceSplitter(model, tokens=chunk_size)
    paper_set = Paperset(dataset, splitter=splitter, content_field=content_field, paragraphs_together=paragraphs)
    api_key = os.getenv("QDRANT_KEY") if key == "QDRANT_KEY" or key == "key" else key
    is_url = "ttp:" in url or "ttps:" in url
    path: Optional[str] = None if is_url else url #actually the user can give either path or url
    url: Optional[str] = url if is_url else None
    logger.info(f"initializing quadrant database at {url}")
    client: QdrantClient = QdrantClient(
        url=url,
        port=6333,
        grpc_port=6334,
        prefer_grpc=is_url if prefer_grpc is None else prefer_grpc,
        api_key=api_key,
        path=path
    )
    client.set_model(model)
    from qdrant_client.http import models as rest
    #client.recreate_collection(collection_name)
    # Just do a single quick embedding to get vector size
    collections = client.get_collections()
    if rewrite or not seq(collections.collections).exists(lambda c: c.name == collection):
        logger.info(f"creating collection {collection}")
        client.recreate_collection(collection_name=collection, vectors_config=client.get_fastembed_vector_params(on_disk=True))
        indexes: dict[str, PayloadSchemaType] = {
            "doi": PayloadSchemaType.TEXT,
            "source": PayloadSchemaType.TEXT,
            "document": PayloadSchemaType.TEXT
        }
        for k, v in indexes.items():
            client.create_payload_index(collection, k, v)
    return paper_set.fast_index_by_slice(n = slice, client=client, collection_name=collection, batch_size=batch_size, parallel=parallel)

@timing
@app.command("dataset")
@click.option('--dataset', type=click.STRING, help="Dataset to index, can be either Path or hugging face dataset")
@click.option('--collection', default='dataset', help='dataset collection name')
@click.option('--url', type=click.STRING, required=True, help="URL or API key for example http://localhost:6333 for qdrant")
@click.option('--key', type=click.STRING, default=None, help="your api key if you are using cloud vector store")
@click.option('--embeddings', type=click.Choice(EMBEDDINGS), default=EmbeddingType.HuggingFaceBGE.value,
              help='embeddings type, huggingface by default')
@click.option('--chunk_size', type=click.INT, default=512, help='size of the chunk for splitting (characters for recursive spliter and tokens for openai one)')
@click.option("--model", type=click.Path(), default=EmbeddingModels.default.value, help="path to the model (required for embeddings)")
@click.option("--device", type=click.Choice(DEVICES), default=Device.cpu.value, help="which device to use, cpu by default, so do not forget to put cuda if you are using NVIDIA")
@click.option('--prefer_grpc', type=click.BOOL, default=False, help = "only needed for qdrant database")
@click.option('--slice', type=click.INT, default=100, help='What is the size of the slice')
@click.option('--start', type=click.INT, default=0, help='When to start slicing the dataset')
@click.option('--content_field', type=click.STRING, default="annotations_paragraph", help = "default dataset content field")
@click.option('--paragraphs', type=click.INT, default=5, help='number of paragraphs to connect together when preprocessing')
@click.option('--rewrite', type=click.BOOL, default=False, help = "Rewrite collection if it is already present")
@click.option('--log_level', type=click.Choice(LOG_LEVELS, case_sensitive=False), default=LogLevel.DEBUG.value, help="logging level")
def index_dataset_command(dataset: str, collection: str, url: Optional[str], key: Optional[str], embeddings: str, chunk_size: int, model: Optional[str], device: Optional[str], prefer_grpc: bool, slice: int, start: int, content_field: str, paragraphs: int, rewrite: bool, log_level: str) -> Path:
    configure_logger(log_level)
    load_environment_keys(usecwd=True)
    assert not (url is None and key is None), "either database url or api_key should be provided!"
    embedding_type = EmbeddingType(embeddings)
    logger.info(f"computing embeddings into collection {collection} for {dataset} of {embeddings} type with model {model} using slices of {slice} starting from {start} with chunks of {chunk_size} tokens when splitting")
    embedding_function = resolve_embeddings(embedding_type, model = model, device = Device(device))
    splitter = resolve_splitter(embedding_type, model, chunk_size)
    paper_set = Paperset(dataset, splitter=splitter, content_field=content_field, paragraphs_together=paragraphs)
    api_key = os.getenv("QDRANT_KEY") if key == "QDRANT_KEY" or key == "key" else key
    indexes: dict[str, PayloadSchemaType] = {
        "doi": PayloadSchemaType.TEXT,
        "source": PayloadSchemaType.TEXT
    }
    db = init_qdrant(collection, path_or_url=url, embeddings=embedding_function, prefer_grpc=prefer_grpc, always_recreate=rewrite, api_key=api_key, indexes=indexes)
    return paper_set.index_by_slices(slice, db, start = start)


if __name__ == '__main__':
    app()


#!/usr/bin/env python3
import functools
import hashlib
import sys
from typing import TypeVar

import loguru
import polars as pl
from beartype import beartype
from datasets import load_dataset
from hybrid_search.opensearch_hybrid_search import OpenSearchHybridSearch
from langchain.schema import Document
from langchain.text_splitter import TextSplitter
from langchain_community.vectorstores import VectorStore, Qdrant
from qdrant_client import QdrantClient

from indexpaper.resolvers import *
from indexpaper.utils import timing

T = TypeVar('T')

DEFAULT_COLUMNS = ('corpusid',
                  'content_source_oainfo_openaccessurl',
                  #'updated',
                  'externalids_doi',
                  'externalids_pubmed',
                  'annotations_abstract',
                  'annotations_author',
                  'annotations_title',
                   'annotations_paragraph',
                   #'content_text'
                   )



class Paperset:
    """
    The class that makes it easier to work with dataset or papers for indexing, can be either hugging face or parquet
    """

    @staticmethod
    def generate_id_from_data(data):
        """
        function to avoid duplicates
        :param data:
        :return:
        """
        if isinstance(data, str):  # check if data is a string
            data = data.encode('utf-8')  # encode the string into bytes
        return str(hex(int.from_bytes(hashlib.sha256(data).digest()[:32], 'little')))[-32:]

    @staticmethod
    @beartype
    def default_transform(content: list[str], size: int = 10, step: int = 10) -> list[str]:
        result = content if len(content) <2 else seq(content).sliding(size, step).map(lambda s: s.reduce(lambda x, y: x + "\n" + y)).to_list()
        #print(f"**************************************************************\nfrom {len(content)} to {len(result)}")
        return result


    lazy_frame: pl.LazyFrame
    content_field: str #"content_text" #'annotations_paragraph'
    splitter: Optional[TextSplitter] = None
    columns: list[str]


    @staticmethod
    def get_dataset(name: str, default_columns: Optional[list|tuple] = DEFAULT_COLUMNS) -> pl.LazyFrame:
        """
        for example "longevity-genie/moskalev_papers"
        :param name:
        :return: polars Dataframe
        """
        dataset = load_dataset(name)["train"]
        df = pl.from_arrow(dataset.data.table).lazy()
        return df if default_columns is None else df.select(default_columns)

    transform_content: Optional[Callable[[list], list]]

    @beartype
    def __init__(self, df_name_or_path: Union[pl.LazyFrame, str, Path],
                 splitter: Optional[TextSplitter] = None,
                 content_field: str = "annotations_paragraph",#'content_text',
                 default_columns=DEFAULT_COLUMNS, low_memory: bool = False,
                 transform_content: Optional[Callable[[list], list]] = None,
                 paragraphs_together: int = 5
                 ):
        self.splitter = splitter
        if isinstance(df_name_or_path, pl.LazyFrame):
            self.lazy_frame = df_name_or_path if default_columns is None else df_name_or_path.select(default_columns)
        elif isinstance(df_name_or_path, Path) or "parquet" in df_name_or_path:
            df = pl.scan_parquet(df_name_or_path, low_memory=low_memory)
            self.lazy_frame = df if default_columns is None else df.select(default_columns)
        else:
            self.lazy_frame = Paperset.get_dataset(df_name_or_path, default_columns)
        self.content_field = content_field
        self.columns = self.lazy_frame.columns
        self.transform_content = functools.partial(self.default_transform, size = paragraphs_together, step = paragraphs_together) if transform_content is None else transform_content
        assert content_field in self.columns, f"{content_field} has not been found in dataframe columns {self.columns}"

    @beartype
    def split_documents(self, documents: list[Document])-> list[Document]:
        """
        splits documents of splitter is present
        :param documents:
        :return:
        """
        return documents if self.splitter is None else self.splitter.split_documents(documents)

    @beartype
    def row_to_documents(self, row: tuple, extras: Optional[Dict] = None):
        """
        Converts dataframe row into the text for indexing
        :param row:
        :param columns:
        :param content_field:
        :param splitter:
        :return:
        """
        d: dict = dict(zip(self.columns, row))
        data = d[self.content_field]
        contents = self.transform_content(data if type(data) is list else [data])
        meta = {k:v for k,v in d.items() if k != self.content_field}
        # Check if extras is not None and update meta with extras
        if extras is not None:
            meta.update(extras)
        def with_index(meta: dict, i: int):
            meta["paragraph"] = i
            if "doi" in meta and meta["doi"] is not None:
                meta["source"] = meta["doi"] + "#" + str(i)
            elif "externalids_doi" in meta and meta["externalids_doi"] is not None:
                meta["doi"] = meta["externalids_doi"]
                meta["source"] = meta["doi"] + "#" + str(i)
            elif "externalids_pubmed" in meta and meta["externalids_pubmed"] is not None:
                meta["source"] = str(meta["externalids_pubmed"]) + "#" + str(i)
            elif "corpusid" in meta and meta["corpusid"] is not None:
                meta["source"] = str(meta["corpusid"]) + "#" + str(i)
            else:
                meta["source"] = "unspecified" + "#" + str(i)
            return meta
        docs: list[Document] = [Document(page_content = c, metadata=with_index(meta, i)) for (c, i) in seq(contents).zip_with_index(1)]
        return self.split_documents(docs)

    @beartype
    def documents_from_dataset_slice(self, df: pl.DataFrame, extras: Optional[Dict] = None) -> list[Document]:
        """
        turns slices of n papers into documents by calling row_to_documeents(r) on each paper
        :param df:
        :return:
        """
        return seq(self.row_to_documents(r, extras) for r in df.iter_rows()).flatten().to_list()


    def fold_left_slices(self, n: int, fold: Callable[[T, pl.DataFrame], T], acc: T, start: int = 0) -> T:
        """
        Function that simulates fold_left on slices of lazy dataframe
        :param df:
        :param n:
        :param fold:
        :param acc:
        :param start:
        :return:
        """
        # Get the slice
        slice_lazy_df = self.lazy_frame.slice(start, n)

        # Collect the slice to a DataFrame to check if it has zero rows
        slice_df = slice_lazy_df.collect()
        if slice_df.shape[0] == 0:
            return acc

        # Apply the function to the slice (in place modification)
        upd_acc = fold(acc, slice_df)

        # Recursive call to process the next slice
        return self.fold_left_slices(n, fold, upd_acc, start + n)

    def fold_left_document_slices(self, n: int, fold: Callable[[T, list[Document]], T], acc: T, start: int = 0) -> T:
        """
        Wrapper to apply it to the documents
        :param n: how many papers to take in a slice
        :param fold:
        :param acc:
        :param start:
        :return:
        """
        def fold_df(value: T, df: pl.DataFrame) -> T:
            return fold(value, self.documents_from_dataset_slice(df))
        return self.fold_left_slices(n, fold_df, acc, start)

    def foreach_numbered_slice(self, n: int, fun: Callable[[pl.DataFrame, int, int], None], start: int = 0) -> None:
        # Get the slice
        slice_lazy_df = self.lazy_frame.slice(start, n)

        # Collect the slice to a DataFrame to check if it has zero rows
        slice_df = slice_lazy_df.collect()
        if slice_df.shape[0] == 0:
            return

        # Apply the function to the slice (in place modification)
        fun(slice_df, n, start)

        # Recursive call to process the next slice
        self.foreach_numbered_slice(n, fun, start + n)

    def foreach_slice(self, n: int, fun: Callable[[pl.DataFrame], None], start: int = 0) -> None:
        # Get the slice
        slice_lazy_df = self.lazy_frame.slice(start, n)

        # Collect the slice to a DataFrame to check if it has zero rows
        slice_df = slice_lazy_df.collect()
        if slice_df.shape[0] == 0:
            return

        # Apply the function to the slice (in place modification)
        fun(slice_df)

        # Recursive call to process the next slice
        self.foreach_slice(n, fun, start + n)

    #@beartype
    @logger.catch(reraise=True)
    def foreach_document_slice(self, n: int, fun: Callable[[list[Document]], None], start: int = 0) -> None:
        def fun_df(df: pl.DataFrame) -> None:
            return fun(self.documents_from_dataset_slice(df))
        sys.setrecursionlimit(100000) #to avoid issues
        result = self.foreach_slice(n, fun_df, start)
        sys.setrecursionlimit(5000) #putting back
        return result

    def foreach_numbered_document_slice(self, n: int, fun: Callable[[list[Document], int, int], None], start: int = 0, extras: Optional[Dict] = None) -> None:
        def fun_df(df: pl.DataFrame, f_n: int, f_start: int) -> None:
            return fun(self.documents_from_dataset_slice(df, extras), f_n, f_start)
        return self.foreach_numbered_slice(n, fun_df, start)

    @beartype
    def fast_index_by_slice(self, n: int, client: QdrantClient, collection_name: str, batch_size: int = 32, start: int = 0, parallel: Optional[int] = None):
        @timing(f"one more slice of {n} papers has been fast-indexed")
        def fast_index_paper_slice(docs: list[Document]) -> None:
            if len(docs) == 0:
                logger.info(f"no more documents to index!")
                return None
            texts = [d.page_content for d in docs]
            for d in docs:
                if "metadata" not in d.metadata: #ugly fix for metadata issue
                    d.metadata["metadata"] = d.metadata.copy()
            metadatas = [d.metadata for d in docs]
            ids = [self.generate_id_from_data(d.page_content) for d in docs]
            client.add(
                collection_name=collection_name,
                documents=texts,
                metadata=metadatas,
                ids=ids,
                batch_size=batch_size,
                parallel=parallel
            )

        return self.foreach_document_slice(n, fast_index_paper_slice, start = start)


    def index_hybrid_by_slices(self, n: int, hybrid: OpenSearchHybridSearch, start: int = 0,
                               pipeline_name: str = "norm-pipeline",
                               logger: Optional["loguru.Logger"] = None,
                               extras: Optional[Dict] = None,
                               verbose: bool = False):
        if not hybrid.check_pipeline_exists(pipeline_name):
            logger.warning(f"hybrid pipeline does not exist, creating pipeline")
            hybrid.create_pipeline(hybrid.opensearch_url, hybrid.login, hybrid.password, pipeline_name)
        log = loguru.Logger if logger is None else logger
        @log.catch(reraise=True)
        @timing(f"one more slice of {n} papers has been indexed")
        def index_paper_slice(docs: list[Document], f_n: int, f_start: int) -> None:
            if len(docs) == 0:
                log.info(f"no more documents to index!")
                return None
            else:
                log.info(f"slice of {f_n} at {f_start} staring from {start}")
            texts = [d.page_content for d in docs]
            metadatas = [d.metadata for d in docs]
            for m in metadatas:  #just technical to track slices
                m["slice"] = f_start
                m["slice_size"] = f_n
            ids = [self.generate_id_from_data(d.page_content) for d in docs]
            hybrid.add_texts(texts=texts, metadatas=metadatas, ids=ids, bulk_size = 15000)

        return self.foreach_numbered_document_slice(n, index_paper_slice, start = start, extras=extras)

    def index_hybrid_by_slices_detailed(self, n: int, index_name: str, embeddings: Embeddings,
                                        url: str,
                                        start: int = 0,
                                        login: Optional[str] = None,
                                        password: Optional[str] = None,
                                        pipeline_name: str = "norm-pipeline",
                                        logger: Optional["loguru.Logger"] = None):
        login = os.getenv("OPENSEARCH_USER", "admin") if login is None else login
        password = os.getenv("OPENSEARCH_PASSWORD", "admin") if password is None else password
        hybrid: OpenSearchHybridSearch = OpenSearchHybridSearch.create(url, index_name, embeddings, login=login, password=password)
        return self.index_hybrid_by_slices(n, hybrid, start, pipeline_name, logger)

    @beartype
    def index_by_slices(self, n: int, db: VectorStore, start: int = 0, logger: Optional["loguru.Logger"] = None):
        """
        :param n: number of papers included in the slice
        :param db: vector store to store results
        :param start: start index
        :param logger: logger to log data to
        :return:
        """
        log = loguru.Logger if logger is None else logger
        @timing(f"one more slice of {n} papers has been indexed")
        def index_paper_slice(docs: list[Document]) -> None:
            if len(docs) == 0:
                log.info(f"no more documents to index!")
                return None
            texts = [d.page_content for d in docs]
            metadatas = [d.metadata for d in docs]
            ids = [self.generate_id_from_data(d.page_content) for d in docs]
            db.add_texts(texts=texts, metadatas=metadatas, ids=ids)

        return self.foreach_document_slice(n, index_paper_slice, start = start)




import os
from enum import Enum

from langchain_community.embeddings import OpenAIEmbeddings, LlamaCppEmbeddings, VertexAIEmbeddings
from langchain_community.embeddings.huggingface import *
from loguru import logger
from pycomfort.files import *

from indexpaper.splitting import OpenAISplitter, SourceTextSplitter, HuggingFaceSplitter


class EmbeddingModels(Enum):
    all_mpnet_base: str = "sentence-transformers/all-mpnet-base-v2"
    bge_large_en: str = "BAAI/bge-large-en" #so far best at https://huggingface.co/spaces/mteb/leaderboard
    bge_base_en_1_5: str = "BAAI/bge-base-en-v1.5" #so far second best at https://huggingface.co/spaces/mteb/leaderboard
    bge_large_en_1_5: str = "BAAI/bge-large-en-v1.5" #so far best at https://huggingface.co/spaces/mteb/leaderboard
    bge_base_en: str = "BAAI/bge-base-en" #so far second best at https://huggingface.co/spaces/mteb/leaderboard

    gte_large: str = "thenlper/gte-large"
    gte_base: str = "thenlper/gte-base"
    multilingual_e5_large: str = "intfloat/multilingual-e5-large" #supports many languages and pretty good
    biobert: str = 'pritamdeka/BioBERT-mnli-snli-scinli-scitail-mednli-stsb'
    pubmedbert: str = 'pritamdeka/S-PubMedBert-MS-MARCO'
    biolord: str = 'FremyCompany/BioLORD-STAMB2-v1' #based on all-mpnet-base-v2 finetined for bio domain
    bioelectra: str = 'menadsa/S-BioELECTRA'
    biolinkbert: str = "michiyasunaga/BioLinkBERT-large" #best on https://microsoft.github.io/BLURB/leaderboard.html
    specter: str = "allenai/specter2_base"
    mxbai_large: str = "mxbai-embed-large-v1"
    default: str = bge_large_en_1_5


class Device(Enum):
    cpu = "cpu"
    cuda = "cuda"
    ipu = "ipu"
    xpu = "xpu"
    mkldnn = "mkldnn"
    opengl = "opengl"
    opencl = "opencl"
    ideep = "ideep"
    hip = "hip"
    ve = "ve"
    fpga = "fpga"
    ort = "ort"
    xla = "xla"
    lazy = "lazy"
    vulkan = "vulkan"
    mps = "mps"
    meta = "meta"
    hpu = "hpu"
    mtia = "mtia"
    privateuseone = "privateuseone"
    cuda0 = "cuda:0"
    cuda1 = "cuda:1"
    cuda2 = "cuda:2"
DEVICES =  [d.value for d in Device]


class VectorDatabase(Enum):
    Chroma = "Chroma"
    Qdrant = "Qdrant"


class EmbeddingType(Enum):
    OpenAI = "openai"
    Llama = "llama"
    VertexAI = "vertexai"
    HuggingFace = "huggingface"
    HuggingFaceBGE = "huggingface_bge"


EMBEDDINGS: list[str] = [e.value for e in EmbeddingType]
VECTOR_DATABASES: list[str] = [db.value for db in VectorDatabase]


def resolve_splitter(embeddings_type: EmbeddingType,

                     model: Optional[Union[Path, str]] = None,
                     chunk_size: Optional[int] = None
                     ) -> SourceTextSplitter:
    """
    initializes a splitter based on embedding type and additional parameters
    :param embeddings_type:
    :param model:
    :param chunk_size:
    :return:
    """
    if embeddings_type == EmbeddingType.OpenAI:
        if chunk_size is None:
            chunk_size = 3600
        return OpenAISplitter(tokens=chunk_size)
    elif embeddings_type==embeddings_type.HuggingFace or embeddings_type==embeddings_type.HuggingFaceBGE:
        if chunk_size is None:
            chunk_size = 512
        if model is None:
            logger.error("Model should be specified for Huggingface splitter, using default sentence-transformers/all-mpnet-base-v2 otherwise")
            return HuggingFaceSplitter("sentence-transformers/all-mpnet-base-v2", tokens=chunk_size)
        else:
            return HuggingFaceSplitter(model, tokens=chunk_size)
    else:
        logger.warning(f"{embeddings_type} splitter is not supported, using openai tiktoken based splitter instead")
        return OpenAISplitter(tokens=chunk_size)


def resolve_embeddings(embeddings_type: EmbeddingType, model: Optional[Union[Path, str]] = None, device: Device = Device.cpu, normalize_embeddings: bool = True) -> Embeddings:
    """
    Initializes embedding instance based on embedding type enum and on additional parameter like model and device
    :param embeddings_type:
    :param model:
    :param device:
    :return:
    """
    if embeddings_type == EmbeddingType.OpenAI:
        return OpenAIEmbeddings()
    elif embeddings_type == EmbeddingType.Llama:
        if model is None:
            model = os.getenv("LLAMA_MODEL")
            if model is None:
                logger.error(f"for llama embeddings for {model} model")
            else:
                return LlamaCppEmbeddings(model_path = model)
        return LlamaCppEmbeddings(model_path = str(model))
    elif embeddings_type == EmbeddingType.VertexAI:
        return VertexAIEmbeddings()
    elif embeddings_type == EmbeddingType.HuggingFace:
        encode_kwargs = {'normalize_embeddings': normalize_embeddings}
        if model is None:
            logger.warning(f"for huggingface the model name should be specified")
            return HuggingFaceEmbeddings(model_kwargs={'device': device.value}, encode_kwargs=encode_kwargs)
        else:
            return HuggingFaceEmbeddings(model_name = str(model), model_kwargs={'device': device.value}, encode_kwargs=encode_kwargs)
    elif embeddings_type == EmbeddingType.HuggingFaceBGE:
        model_name = "BAAI/bge-large-en" if model is None else model
        model_kwargs = {'device': device.value}
        encode_kwargs = {'normalize_embeddings': normalize_embeddings} # set True to compute cosine similarity
        return HuggingFaceBgeEmbeddings(
            model_name=model_name,
            model_kwargs=model_kwargs,
            encode_kwargs=encode_kwargs
        )
    else:
        logger.warning(f"{embeddings_type.value} is not yet supported by CLI, using default openai embeddings instead")
        return OpenAIEmbeddings()

def resolve_embedding_splitter(embeddings_type: EmbeddingType, model: Optional[Union[Path, str]] = None, device: Device = Device.cpu, chunk_size=512):
    return resolve_embeddings(embeddings_type, model, device), resolve_splitter(embeddings_type, model, chunk_size=chunk_size)



import json
from abc import ABC
from copy import deepcopy
from typing import Dict, Union, List

import tiktoken
from langchain.schema import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from loguru import logger
from pycomfort.files import *
from transformers import AutoTokenizer
from transformers import PreTrainedTokenizerBase

def flatten_dict(d: Dict[str, Union[str, int, float, Dict, List]], parent_key: str = '', sep: str = '_', list_sep: Optional[str] = ", ") -> Dict[str, Union[str, int, float]]:
    """
    Flattens a nested dictionary, converting list values to comma-separated strings, and ignoring None values.

    :param d: The dictionary to be flattened
    :param parent_key: The key from the parent dictionary, if any
    :param sep: The separator to be used between parent and child keys
    :param list_sep: The separator to be used when joining list items into a string, if None - just drops the lists
    :return: The flattened dictionary
    """

    items: Dict[str, Union[str, int, float]] = {}  # Initialize empty dictionary to hold result

    for k, v in d.items():  # Iterate over items in input dictionary
        new_key = f"{parent_key}{sep}{k}" if parent_key else k  # Generate new key

        if isinstance(v, dict):  # If value is a dictionary, flatten it recursively
            items.update(flatten_dict(v, new_key, sep=sep, list_sep=list_sep))

        elif isinstance(v, list):  # If value is a list, convert to string
            if list_sep is not None:
                list_str = "[ " + list_sep.join(map(str, v)) + " ]"
                if list_str is not None:  # Ignore None values
                    items[new_key] = list_str

        elif v is not None:  # Ignore None values
            items[new_key] = v  # If value is not a dictionary or a list, just add to result

    return items  # Return the flattened dictionary


def paginated_paper_to_documents(folder: Path) -> list[Document]:
    metas: list[Path] = [folder] if "_meta.json" in folder.name else traverse(folder, lambda p: "_meta.json" in p.name)
    docs = []
    for meta in metas:
        parent = meta.parent
        json_data = meta.read_text("utf-8")
        meta_dic = json.loads(json_data)
        paper = parent / meta.name.replace("_meta.json", "")
        if not paper.exists():
            print(f"cannot find paper for {meta}")
        else:
            sorted_files = sorted(paper.glob("*.txt"), key=lambda x: int(x.stem.split('_')[-1]))
            for i, text in enumerate(sorted_files):
                doc = paper_to_document(text, meta_dic, extra={"page": i})
                if doc is not None:
                    docs.append(doc)
    return docs



def papers_to_documents(folder: Path, suffix: str = "", include_meta: bool = True):
    txt = traverse(folder, lambda p: "txt" in p.suffix)
    texts = [t for t in txt if suffix in t.name] if suffix != "" else txt
    docs: List[Document] = []
    for t in texts:
        meta: Path = t.parent / t.name.replace(".txt", "_meta.json")
        if include_meta and meta.exists:
            json_data = meta.read_text("utf-8")
            meta_dic = json.loads(json_data)
            doc = paper_to_document(t, meta_dic)
        else:
            doc = paper_to_document(t)
        if doc is not None:
            docs.append(doc)
    return docs


def paper_to_document(paper: Path, meta: Optional[dict] = None, min_tokens: int = 100, extra: Optional[dict] = None) -> Optional[Document]:
    """
    Turns paper into document, assumes the folder/paper_name is DOI
    :param paper:
    :param meta: additional metadata
    :param min_tokens:
    :param extra: optional extra fields to add
    :return: Documenty
    """
    doi = f"http://doi.org/{paper.parent.name}/{paper.stem}"
    text = paper.read_text(encoding="utf-8")
    if len(text) < min_tokens:
        logger.warning("TOO SHORT TEXT")
        return None
    else:
        new_meta = {} if meta is None else flatten_dict(meta)
        new_meta["source"] = doi
        new_meta["doi"] = doi
        if extra is not None:
            for k, v in extra.items():
                new_meta[k] = v
        return Document(
            page_content=text,
            metadata=new_meta
        )


class SourceTextSplitter(RecursiveCharacterTextSplitter, ABC):
    """
    Class that insludes dois and paging into metadata
    """

    @property
    def chunk_size(self):
        return self._chunk_size

    def create_documents(
            self, texts: List[str], metadatas: Optional[List[dict]] = None
    ) -> List[Document]:
        """Create documents from a list of texts."""
        _metadatas = metadatas or [{}] * len(texts)
        documents = []
        for i, text in enumerate(texts):
            meta = _metadatas[i]
            source: Optional[str] = meta["source"] if "source" in meta else None
            #if "paragraph" in meta and meta["paragraph"] is not None:
            #    source = source + "#paragraph_" + str(meta["paragraph"])
            sub_texts = self.split_text(text)

            for j, chunk in enumerate(sub_texts, start=1):
                new_meta = deepcopy(meta)
                if source is not None:
                    num = str(j)
                    if len(sub_texts) > 1:
                        new_meta["source"] = source + "." + num if "#" in source else source + "#" + num
                    if "doi" not in new_meta:
                        new_meta["doi"] = source
                    if "split" not in new_meta:
                        new_meta["split"] = num
                new_doc = Document(page_content=chunk, metadata=new_meta)
                documents.append(new_doc)
        return documents


class HuggingFaceSplitter(SourceTextSplitter, ABC):

    tokenizer: PreTrainedTokenizerBase

    def __init__(self,
                 model: str = "sentence-transformers/all-mpnet-base-v2", #"thenlper/gte-large"
                 tokens: int = 512, #based on benchmarks
                 tokens_overlap: int = 0,
                 keep_separator: bool = False,
                 add_start_index: bool = False
                 ):

        self.tokenizer = AutoTokenizer.from_pretrained(model)

        def length_function(text: str) -> int:
            return len(self.tokenizer.encode(text, padding=True, truncation=True, return_tensors='pt'))

        super().__init__(chunk_size=tokens, chunk_overlap=tokens_overlap,
                         length_function=length_function,
                         keep_separator=keep_separator,
                         add_start_index=add_start_index)



class OpenAISplitter(SourceTextSplitter, ABC):

    def __init__(self, model: str = "gpt-3.5-turbo-16k",
                 tokens: int = 2000,
                 tokens_overlap: int = 0,
                 keep_separator: bool = False,
                 add_start_index: bool = False
                 ):

        def length_function(text: str) -> int:
            encoding = tiktoken.encoding_for_model(model)
            return len(encoding.encode(text))

        super().__init__(chunk_size=tokens, chunk_overlap=tokens_overlap,
                   length_function=length_function,
                   keep_separator=keep_separator,
                   add_start_index=add_start_index)



from typing import Optional

import click
from hybrid_search.opensearch_hybrid_search import OpenSearchHybridSearch
from langchain_community.embeddings import HuggingFaceBgeEmbeddings


@click.group(invoke_without_command=True)
@click.pass_context
def app(ctx):
    if ctx.invoked_subcommand is None:
        ctx.invoke(hybrid)

@app.command("hybrid")
@click.option('--url', default='http://localhost:9200', help='URL of the OpenSearch instance')
@click.option('--index', help='Name of the index in OpenSearch')
@click.option('--device', default='cpu', help='Device to run the model on (e.g., cpu, cuda)')
@click.option('--model', default='BAAI/bge-large-en-v1.5', help='Name of the model to use')
@click.option('--query', default='What is ageing?', help='The query to search for')
@click.option('--k', default=10, help='Number of search results to return')
@click.option('--threshold', default=None, help='Threshold to cut out results')
@click.option('--verbose', default=True, help='How much to print')
def hybrid(url: str, index: str, device: str, model: str, query: str, k: int, threshold: Optional[float], verbose: bool):
    print(f"searching in INDEX: {index}, \nQUERY: {query}")
    model_kwargs = {"device": device}
    encode_kwargs = {"normalize_embeddings": True}

    embeddings = HuggingFaceBgeEmbeddings(
        model_name=model,
        model_kwargs=model_kwargs,
        encode_kwargs=encode_kwargs
    )

    docsearch: OpenSearchHybridSearch = OpenSearchHybridSearch.create(url, index, embeddings)
    results = docsearch.hybrid_search(query, k, search_pipeline = "norm-pipeline", threshold=threshold)
    print("Search IDS:")
    for (result, f) in results:
        print(result.metadata["source"], f)
        if verbose:
            print(result)



if __name__ == '__main__':
    app()


import json
from typing import Optional
import click
import requests
from pathlib import Path

# Create a context class to hold global options
class Context:
    def __init__(self, key: str):
        self.key = key

# Setup a group with a shared context
@click.group()
@click.option('--key', type=str, help='Semantic Scholar API key', required=True)
@click.pass_context
def cli(ctx: click.Context, key: str):
    """CLI tool to interact with Semantic Scholar's dataset API."""
    ctx.obj = Context(key)

def make_request(url: str, ctx: click.Context) -> dict:
    headers = {
        'x-api-key': ctx.obj.key,
        'Accept': 'application/json'  # Explicitly request JSON response
    }
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    return response.json()

def write_output(data: dict, output_path: Optional[Path]):
    if output_path:
        output_path.write_text(json.dumps(data, indent=2))
        click.echo(f"Output written to {output_path}")
    else:
        click.echo(data)

@cli.command()
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), help='File path to write the output JSON.')
def last_three_releases(ctx: click.Context, output: Optional[click.Path]):
    """Fetch and print the last three release dates."""
    url = 'https://api.semanticscholar.org/datasets/v1/release'
    releases = make_request(url, ctx)
    write_output(releases[-3:], Path(output) if output else None)

@cli.command("latest_release")
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), help='File path to write the output JSON.')
def latest_release(ctx: click.Context, output: Optional[click.Path]):
    """Fetch and print the latest release ID and details of the first dataset."""
    url = 'https://api.semanticscholar.org/datasets/v1/release/latest'
    latest = make_request(url, ctx)
    write_output(latest, Path(output) if output else None)

@cli.command("dataset_info")
@click.argument('dataset_name', type=str)
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), help='File path to write the output JSON.')
def dataset_info(ctx: click.Context, dataset_name: str, output: Optional[click.Path]):
    """Fetch and print info for a specific dataset in the latest release."""
    url = f'https://api.semanticscholar.org/datasets/v1/release/latest/dataset/{dataset_name}'
    dataset_info = make_request(url, ctx)
    write_output(dataset_info, Path(output) if output else None)

@cli.command("s2orc")
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), help='File path to write the output JSON.')
@click.option('--field', type=str, required = False, default = None, help='File path to write the output JSON.')
def s2orc(ctx: click.Context, output: Optional[click.Path], field: str):
    """Fetch and print info for a specific dataset in the latest release."""
    url = f'https://api.semanticscholar.org/datasets/v1/release/latest/dataset/s2orc'
    dataset_info = make_request(url, ctx)
    print(f"FIELD IS "+field)
    result = dataset_info if field is None else dataset_info[field]
    write_output(result, Path(output) if output else None)

@cli.command("s2orc_files")
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), default= "s2orc_files.txt", help='File path to write the output')
def s2orc(ctx: click.Context, output: Optional[click.Path]):
    """Fetch and print info for a specific dataset in the latest release."""
    url = f'https://api.semanticscholar.org/datasets/v1/release/latest/dataset/s2orc'
    dataset_info = make_request(url, ctx)
    filenames = dataset_info["files"]
    with Path(output).open('w') as file:
        # Iterate through each filename in the array
        for filename in filenames:
            # Write the filename to the file followed by a newline
            file.write(filename + '\n')

@cli.command("incremental_updates")
@click.argument('dataset_name', type=str)
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), help='File path to write the output JSON.')
def incremental_updates(ctx: click.Context, dataset_name: str, output: Optional[click.Path]):
    """Fetch and print incremental updates for a specific dataset."""
    url = f'https://api.semanticscholar.org/datasets/v1/dataset/{dataset_name}/updates'
    updates = make_request(url, ctx)
    write_output(updates, Path(output) if output else None)

@cli.command("release_data")
@click.argument('release_id', type=str)
@click.argument('dataset_name', type=str)
@click.pass_context
@click.option('--output', type=click.Path(dir_okay=False, writable=True), help='File path to write the output JSON.')
def release_data(ctx: click.Context, release_id: str, dataset_name: str, output: Optional[click.Path]):
    """Fetch and print detailed release data for a specific dataset within a specified release."""
    url = f'https://api.semanticscholar.org/datasets/v1/release/{release_id}/dataset/{dataset_name}'
    release_data = make_request(url, ctx)
    write_output(release_data, Path(output) if output else None)

if __name__ == '__main__':
    cli(obj={})



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.23'
DESCRIPTION = 'indexpaper - library to index papers with vector databases'
LONG_DESCRIPTION = 'indexpaper - library to index papers with vector databases'

# Setting up
setup(
    name="indexpaper",
    version=VERSION,
    author="antonkulaga (Anton Kulaga)",
    author_email="<antonkulaga@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['pycomfort>=0.0.15', 'hybrid-search>=0.0.12', 'more-itertools', 'click', 'python-dotenv', 'tiktoken', 'FlagEmbedding>=1.2.8',
                      'langchain>=0.1.12', "langchain-community>=0.0.28", 'openai', 'Deprecated', 'loguru', '',#'fastembed>=0.2.5',
                      'qdrant-client>=1.8.0', 'sentence_transformers', 'datasets', 'polars>=0.20.14', 'beartype'],
    keywords=['python', 'utils', 'files', 'papers', 'download', 'index', 'vector databases'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
     "console_scripts": [
         "index=indexpaper.index:app"
     ]
    }
)



#!/usr/bin/env python3
import hashlib
import time
from typing import Any

import click
import langchain
from beartype import beartype
from click import Context
from langchain.schema import Document
from langchain.text_splitter import TextSplitter
from langchain_community.vectorstores import Chroma, VectorStore, Qdrant
from pycomfort.config import load_environment_keys
from qdrant_client import QdrantClient
from qdrant_client.http.models import PayloadSchemaType

from indexpaper.resolvers import *
from indexpaper.splitting import SourceTextSplitter, papers_to_documents, paginated_paper_to_documents
from indexpaper.utils import timing


def generate_id_from_data(data):
    """
    function to avoid duplicates
    :param data:
    :return:
    """
    if isinstance(data, str):  # check if data is a string
        data = data.encode('utf-8')  # encode the string into bytes
    return str(hex(int.from_bytes(hashlib.sha256(data).digest()[:32], 'little')))[-32:]

def db_with_texts(db: VectorStore, texts: list[str],
                  splitter: TextSplitter, id_field: Optional[str] = None, debug: bool = False):
    return db_with_documents(db, texts_to_documents(texts), splitter, id_field, debug)


def db_with_documents(db: VectorStore, documents: list[Document],
                      splitter: TextSplitter,
                      id_field: Optional[str] = None, debug: bool = False):
    """
    Function to add documents to a Chroma database.

    Args:
        db (Chroma): The database to add the documents to.
        documents (list[Document]): The list of documents to add.
        splitter (TextSplitter): The TextSplitter to use for splitting the documents.
        debug (bool): If True, print debug information. Default is False.
        id_field (Optional[str]): If provided, use this field from the document metadata as the ID. Default is None.

    Returns:
        Chroma: The updated database.
    """
    docs = splitter.split_documents(documents)
    texts = [doc.page_content for doc in docs]
    metadatas = [doc.metadata for doc in docs]
    ids = [doc.metadata[id_field] for doc in docs] if id_field is not None else None
    if debug:
        for doc in documents:
            logger.trace(f"ADD TEXT: {doc.page_content}")
            logger.trace(f"ADD METADATA {doc.metadata}")
    db.add_texts(texts=texts, metadatas=metadatas, ids=ids)
    return db

@beartype
def init_qdrant(collection_name: str,
                path_or_url: Optional[str],
                embeddings: Optional[Embeddings],
                always_recreate: bool = True,
                api_key: Optional[str] = None,
                distance_func: str = "Cosine",
                prefer_grpc: bool = False,
                timeout: Optional[int] = 3600,
                indexes: Optional[dict[str, PayloadSchemaType]] = None
                ):
    """
    Function that initializes QDrant
    :param collection_name:
    :param path_or_url:
    :param embedding_function:
    :param api_key:
    :param distance_func:
    :param prefer_grpc:
    :return:
    """
    is_url = "ttp:" in path_or_url or "ttps:" in path_or_url
    path: Optional[str] = None if is_url else path_or_url
    url: Optional[str] = path_or_url if is_url else None
    logger.info(f"initializing quadrant database at {path_or_url}")
    client: QdrantClient = QdrantClient(
        url=url,
        port=6333,
        grpc_port=6334,
        prefer_grpc=is_url if prefer_grpc is None else prefer_grpc,
        api_key=api_key,
        path=path
    )
    from qdrant_client.http import models as rest
    #client.recreate_collection(collection_name)
    # Just do a single quick embedding to get vector size
    collections = client.get_collections()
    if always_recreate or not seq(collections.collections).exists(lambda c: c.name == collection_name):
        partial_embeddings = embeddings.embed_documents(["Hello world text!"])
        vector_size = len(partial_embeddings[0])
        logger.info(f"creating collection {collection_name},\n computed probe vector size for the model was {vector_size}")
        distance_func = distance_func.upper()
        client.recreate_collection(
            collection_name=collection_name,
            vectors_config=rest.VectorParams(
                size=vector_size,
                distance=rest.Distance[distance_func]
            ), timeout=timeout
        )
        for k, v in indexes.items():
            client.create_payload_index(collection_name, k, v)
    return Qdrant(client, collection_name=collection_name, embeddings=embeddings)


def write_remote_db(url: str,
                    collection_name: str,
                    documents: list[Document],
                    splitter: TextSplitter,
                    id_field: Optional[str] = None,
                    embeddings: Optional[Embeddings] = None,
                    database: VectorDatabase = VectorDatabase.Qdrant,
                    key: Optional[str] = None,
                    prefer_grpc: Optional[bool] = False,
                    always_recreate: bool = False) -> (Union[VectorStore, Any, Chroma], Optional[Union[Path, str]], float):
    if database == VectorDatabase.Qdrant:
        logger.info(f"writing a collection {collection_name} of {len(documents)} documents to quadrant db at {url}")
        start_time = time.perf_counter()
        api_key = os.getenv("QDRANT_KEY") if key == "QDRANT_KEY" or key == "key" else key
        db = init_qdrant(collection_name, path_or_url=url, embeddings=embeddings, api_key=api_key, prefer_grpc=prefer_grpc, always_recreate=always_recreate)
        db_updated = db_with_documents(db, documents, splitter,  id_field)
        end_time = time.perf_counter()
        execution_time: float = end_time - start_time
        return db_updated, url, execution_time
    else:
        raise Exception(f"Remote Chroma is not yet supported by this script!")
    pass


@beartype
def make_local_db(collection_name: str,
                  documents: list[Document],
                  splitter: TextSplitter,
                  embeddings: Optional[Embeddings] = None,
                  database: VectorDatabase = VectorDatabase.Chroma,
                  persist_directory: Optional[Path] = None,
                  id_field: Optional[str] = None,
                  prefer_grpc: Optional[bool] = False,
                  always_recreate: bool = False
                  ) -> (Union[VectorStore, Any, Chroma], Optional[Union[Path, str]], float):
    """
    :param collection_name:
    :param documents:
    :param splitter:
    :param embeddings:
    :param database:
    :param persist_directory:
    :param id_field:
    :param prefer_grpc:
    :return:
    """
    start_time = time.perf_counter()

    # If no embeddings were provided, default to OpenAIEmbeddings
    if embeddings is None:
        embeddings = OpenAIEmbeddings()

    # Create the directory where the database will be saved, if it doesn't already exist
    if persist_directory is not None:
        where = persist_directory / collection_name
        where.mkdir(exist_ok=True, parents=True)
        where_str = str(where)
    else:
        where = None
        where_str = None

    # Create a Chroma database with the specified collection name and embeddings, and save it in the specified directory
    if database == VectorDatabase.Qdrant:
        db = init_qdrant(collection_name, where_str, embedding_function=embeddings,  prefer_grpc = prefer_grpc, always_recreate = always_recreate)
    else:
        db = Chroma(collection_name=collection_name, persist_directory=persist_directory, embedding_function=embeddings)
    db_updated = db_with_documents(db, documents, splitter,  id_field)

    # Persist the changes to the database
    if persist_directory is not None:
        db_updated.persist()
    end_time = time.perf_counter()
    execution_time: float = end_time - start_time

    # Return the directory where the database was saved
    return db_updated, where, execution_time


@click.group(invoke_without_command=False)
@click.pass_context
def app(ctx: Context):
    # if ctx.invoked_subcommand is None:
    #    click.echo('Running the default command...')
    #    test_index()
    pass


@beartype
def texts_to_documents(texts: list[str]) -> list[Document]:
    return [Document(
        page_content=text,
        metadata={"text": text}
    ) for text in texts]

@timing("indexing selected")
def index_selected_documents(documents: list[Document],
                          collection: str,
                          splitter: SourceTextSplitter,
                          embedding_type: EmbeddingType,
                          url: str,
                          key: str,
                          database: VectorDatabase = VectorDatabase.Chroma.value,
                          model: Optional[Union[Path, str]] = None,
                          prefer_grpc: Optional[bool] = None,
                          device: Device = Device.cpu
                          ) -> (Union[VectorStore, Any, Chroma], Optional[Union[Path, str]], float):
    openai_key = load_environment_keys() #for openai key
    embeddings_function = resolve_embeddings(embedding_type, model, device)
    logger.info(f"embeddings are {embedding_type}")
    return write_remote_db(url, collection, documents, splitter, embeddings=embeddings_function, database=database, key=key, prefer_grpc=prefer_grpc)


@beartype
def index_selected_papers(folder_or_texts: Union[Path, list[str]],
                          collection: str,
                          splitter: SourceTextSplitter,
                          embedding_type: EmbeddingType,
                          include_meta: bool = True,
                          folder: Optional[Union[Path,str]] = None,
                          url: Optional[str] = None,
                          key: Optional[str] = None,
                          database: VectorDatabase = VectorDatabase.Chroma.value,
                          model: Optional[Union[Path, str]] = None,
                          prefer_grpc: Optional[bool] = None,
                          always_recreate: bool = False,
                          device: Device = Device.cpu
                          ) -> (Union[VectorStore, Any, Chroma], Optional[Union[Path, str]], float):
    openai_key = load_environment_keys() #for openai key
    embeddings_function = resolve_embeddings(embedding_type, model, device)
    logger.info(f"embeddings are {embedding_type}")
    documents = papers_to_documents(folder_or_texts, include_meta=include_meta) if isinstance(folder_or_texts, Path) else texts_to_documents(folder_or_texts)
    if url is not None or key is not None:
        return write_remote_db(url, collection, documents, splitter, embeddings=embeddings_function, database=database, key=key, prefer_grpc = prefer_grpc, always_recreate = always_recreate)
    else:
        if folder is None:
            logger.warning(f"neither url not folder are set, trying in memory")
            return make_local_db(collection, documents, splitter, embeddings_function, prefer_grpc = prefer_grpc, database=database, always_recreate = always_recreate)
        else:
            index = Path(folder) if isinstance(folder, str) else folder
            index.mkdir(exist_ok=True)
            where = index / f"{embedding_type.value}_{splitter.chunk_size}_chunk"
            where.mkdir(exist_ok=True, parents=True)
            logger.info(f"writing index of papers to {where}")
            return make_local_db(collection, documents, splitter, embeddings_function, persist_directory=where,  prefer_grpc = prefer_grpc, database=database, always_recreate = always_recreate)


def fast_index_papers(folder: Path, collection: str, url: Optional[str], key: Optional[str], model: str = EmbeddingModels.default.value, prefer_grpc: bool = True, parallel: Optional[int] = None, rewrite: bool = False, paginated: bool = True) -> Path:
    load_environment_keys(usecwd=True)
    assert not (url is None and key is None), "either database url or api_key should be provided!"
    chunk_size: int = 512
    #logger.info(f"computing embeddings into collection {collection} for {dataset} with model {model} using slices of {slice} starting from {start} with chunks of {chunk_size} tokens when splitting")
    splitter = HuggingFaceSplitter(model, tokens=chunk_size)
    api_key = os.getenv("QDRANT_KEY") if key == "QDRANT_KEY" or key == "key" else key
    is_url = "ttp:" in url or "ttps:" in url
    path: Optional[str] = None if is_url else url #actually the user can give either path or url
    url: Optional[str] = url if is_url else None
    logger.info(f"initializing quadrant database at {url}")
    client: QdrantClient = QdrantClient(
        url=url,
        port=6333,
        grpc_port=6334,
        prefer_grpc=is_url if prefer_grpc is None else prefer_grpc,
        api_key=api_key,
        path=path
    )
    client.set_model(model)
    collections = client.get_collections()
    if rewrite or not seq(collections.collections).exists(lambda c: c.name == collection):
        logger.info(f"creating collection {collection}")
        client.recreate_collection(collection_name=collection, vectors_config=client.get_fastembed_vector_params(on_disk=True))
        indexes: dict[str, PayloadSchemaType] = {
            "doi": PayloadSchemaType.TEXT,
            "source": PayloadSchemaType.TEXT,
            "document": PayloadSchemaType.TEXT
        }
        for k, v in indexes.items():
            client.create_payload_index(collection, k, v)
    docs = paginated_paper_to_documents(folder) if paginated else papers_to_documents(folder)
    texts = [d.page_content for d in docs]
    for d in docs:
        if "metadata" not in d.metadata: #ugly fix for metadata issue
            d.metadata["metadata"] = d.metadata.copy()
    metadatas = [d.metadata for d in docs]
    ids = [generate_id_from_data(d.page_content) for d in docs]
    client.add(
        collection_name=collection,
        documents=texts,
        metadata=metadatas,
        ids=ids,
        parallel=parallel
    )
    logger.info(f"added {len(texts)} documents")
    return client
    #paper_set.fast_index_by_slice(n = slice, client=client, collection_name=collection, batch_size=batch_size, parallel=parallel)